# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Any, Dict, Literal, Optional, List, Union
from datetime import datetime, timezone
from pydantic import Field, model_validator

from neon_data_models.enum import SubmindStatus, CcaiState, CcaiControl
from neon_data_models.types import BotType
from neon_data_models.models.api.llm import LLMPersona
from neon_data_models.models.api.chatbots import ConnectedSubmind
from neon_data_models.models.base import BaseModel
from neon_data_models.models.base.contexts import KlatContext, MQContext


class ChatbotsMqRequest(KlatContext, MQContext):
    """
    Defines a request from Klat to the Chatbots service.
    """
    username: str = Field(alias="nick",
                          description="Username (or 'nick') of the sender")
    cid: str = Field(description="Conversation ID associated with the shout")
    message_text: str = Field(alias="messageText",
                              description="Text content of the shout")
    from_bot: bool = Field(
        default=False,
        description="True if the shout is from a bot, False if from a user")
    prompt_id: Optional[str] = Field(
        default=None,
        description="ID of the CCAI prompt associated with the shout")
    prompt_state: Optional[CcaiState] = Field(
        default=None, deprecated=True,
        description="State of the CCAI conversation associated with the shout")
    time_created: datetime = Field(
        default= datetime.now(tz=timezone.utc),
        description="Timestamp when the shout was created")
    requested_participants: Optional[List[str]] = Field(
        default=None, alias="participating_subminds",
        description="List of CCAI participants requested to handle the shout")
    recipient: Optional[str] = Field(
        default=None, description="Explicitly defined recipient of the shout")
    bound_service: Optional[str] = Field(
        default=None, description="Service bound to the conversation")
    context: Optional[dict] = Field(
            default=None, deprecated=True, description="Extra proctor context")
    
    @classmethod
    def from_sio_message(cls, sio_message: dict) -> 'ChatbotsMqRequest':
        klat_context = KlatContext(**sio_message)
        mq_context = MQContext(**sio_message)
        return ChatbotsMqRequest(
            **klat_context.model_dump(exclude_none=True),
            **mq_context.model_dump(exclude_none=True),
            username=sio_message.get("userDisplayName") or \
                sio_message.get("userID"),
            message_text=sio_message["messageText"],
            from_bot=sio_message.get("bot") == 1,
            prompt_id = sio_message.get("promptID"),
            prompt_state=sio_message.get("promptState"),
            time_created=sio_message["timeCreated"],
            recipient=sio_message.get("recipient"),
            bound_service=sio_message.get("bound_service"),
        )

    def model_dump(self, **kwargs):
        """
        Override model_dump to include SIO fields for backwards compatibility
        """

        # For backwards-compat with Klat Server, include aliased keys in 
        # serialization. In the future, this should be configurable and
        # eventually removed.
        by_alias = {}
        if 'by_alias' not in kwargs:
            by_alias = super().model_dump(by_alias=True, **kwargs)

        # Add parameters for backwards-compat.
        by_alias["bot"] = "1" if self.from_bot else "0"

        return {**super().model_dump(**kwargs), **by_alias}


class ChatbotsMqSubmindResponse(KlatContext, MQContext):
    """
    Defines a chatbot response to a request.
    """
    user_id: str = Field(alias='userID', 
                         description="Unique UID of the sender")
    username: Optional[str] = Field(default=None,
                                    alias="userDisplayName",
                                    description="Username of the sender")
    message_text: str = Field(alias="messageText",
                              description="Text content of the shout")
    sid: str = Field(default="", alias="messageID", description="Shout ID")
    replied_message: Optional[str] = Field(
        default=None, alias="repliedMessage",
        description="ID of the shout being replied to")
    bot: Literal["0", "1"] = Field(default='0', alias='is_bot',
                                   description="1 if the shout is from a bot")
    prompt_id: Optional[str] = Field(
        default=None, alias="promptID",
        description="ID of the CCAI prompt associated with the shout")
    is_announcement: bool = Field(
        default=False, alias="isAnnouncement",
        description="True if the shout is an announcement")
    time_created: datetime = Field(
        default= datetime.now(tz=timezone.utc), alias="timeCreated",
        description="Timestamp when the shout was created")
    source: str = Field(
        default="klat_observer",
        description="Name of the service originating the shout")
    bot_type: Optional[BotType] = Field(default=None, deprecated=True,
                              description="Type of submind sending the shout")
    
    # Below are deprecated fields for backwards-compat.
    service_name: Any = Field(default=None, deprecated=True)
    context: Optional[dict] = Field(
        default=None, deprecated=True,
        description="Context used for Klat Server backwards-compat.")
    dom: Any = Field(default=None, deprecated=True,
                     description="Domain of this conversation")
    omit_reply: bool = Field(
        default=True, deprecated=True,
        description="If true, the Proctor will ignore this message")
    no_save: bool = Field(default=False, deprecated=True,
                          description="If true, this message will be ignored")
    to_discussion: bool = Field(default=False, deprecated=True)
    prompt_state: CcaiState = Field(
        default=CcaiState.IDLE, deprecated=True, alias="promptState",
        description="State of the CCAI conversation associated with the shout")
    
    @model_validator(mode='after')
    def set_username_from_user_id(self):
        if self.username is None and self.user_id:
            self.username = self.user_id.rsplit('-', 1)[0]
        return self

    @model_validator(mode='before')
    @classmethod
    def validate_inputs(cls, values):
        if isinstance(values, dict):
            # Some additional aliases for backwards-compat.
            if "nick" in values:
                values.setdefault("userID", values.get("nick"))

            if "shout" in values:
                values.setdefault("messageText", values.get("shout"))
            
            if "responded_shout" in values:
                values.setdefault("repliedMessage",
                                   values.get("responded_shout"))

            if "time" in values:
                values.setdefault("timeCreated", values.get("time"))

            if "created_on" in values:
                values.setdefault("timeCreated", values.get("created_on"))

            if "sid" in values and values["sid"] is None:
                values.pop("sid")

            if "conversation_state" in values:
                values.setdefault("promptState",
                                  values.get("conversation_state"))

        return values

    def model_dump(self, **kwargs):
        # For backwards-compat with Klat Server, include aliased keys in 
        # serialization. In the future, this should be configurable and
        # eventually removed.
        by_alias = {}
        if 'by_alias' not in kwargs:
            by_alias = super().model_dump(by_alias=True, **kwargs)

        by_alias['isAnnouncement'] = '1' if self.is_announcement else '0'
        by_alias['nick'] = self.user_id
        by_alias['responded_shout'] = self.replied_message
        by_alias['shout'] = self.message_text
        by_alias['time'] = self.time_created.timestamp()
        by_alias['promptState'] = self.prompt_state.value
        by_alias['created_on'] = self.time_created.timestamp()
        
        return {**super().model_dump(**kwargs), **by_alias}


class PromptCompletedContext(BaseModel):
    prompt: Optional[ChatbotsMqRequest] = Field(
        default=None, description="Original request containing the prompt")

    prompt_text: str = Field(description="The string prompt that is completed")
    available_subminds: List[str] = Field(  # Seems to always match `participating_subminds`
        default=[], description="List of subminds available to participate")
    participating_subminds: List[str] = Field(
        default=[], description="List of subminds participating in the prompt")
    proposed_responses: Dict[str, str] = Field(
        default={}, description="Dict of nick to proposal")
    
    # In the future, there will be a list of these for multi-round discussion
    submind_opinions: Dict[str, str] = Field(
        default={}, description="Dict of nick to discussion")

    votes: Dict[str, str] = Field(default={},
                                  description="Dict of nick to vote")
    votes_per_submind: Dict[str, List[str]] = Field(
        default={}, description="Dict of nick to list of received votes")
    winner: str = Field(default="", description="Selected winner")

    # Below are deprecated
    is_active: bool = Field(  # Seems to report active all the time
        default=False, deprecated=True,
        description="True if a response has not yet been chosen")
    state: Optional[CcaiState] = Field(
        default=CcaiState.PICK, deprecated=True,
        description="State of the CCAI conversation (always PICK)")


class ChatbotsMqSavePrompt(ChatbotsMqSubmindResponse):
    context: PromptCompletedContext = Field(
        alias="conversation_context",
        description="Definition of the completed discussion")
    
    def model_dump(self, **kwargs):
        return ChatbotsMqSubmindResponse.model_dump(self, **kwargs)


class ChatbotsMqNewPrompt(ChatbotsMqSubmindResponse):
    prompt_id: str = Field(
        description="ID of the CCAI prompt associated with the shout"
    )
    user_id: Optional[str] = Field(default=None, alias="nick",
                                   description="User ID of the proctor")
    prompt_text: str = Field(description="The new prompt being discussed")
    prompt_state: CcaiState = Field(
        default=CcaiState.IDLE, deprecated=True,
        description="Implemented for backwards-compat. New Prompt always IDLE")
    discussion_rounds: int = Field(
        default=2, 
        description="Number of discussion rounds per cycle for this prompt")
    context: Optional[dict] = Field(default=None, deprecated=True,
                                    alias="conversation_context")

    @model_validator(mode='before')
    @classmethod
    def validate_context(cls, values):
        values.setdefault("message_text", values.get("shout", ""))
        values.setdefault("user_id", values.get("nick"))
        return values
    
    def model_dump(self, **kwargs):
        return ChatbotsMqSubmindResponse.model_dump(self, **kwargs)


class ChatbotsMqResponse:
    """
    Type adapter for validating an arbitrary MQ message. This will always return
    an instance that extends `BaseMessage` and `MQContext`.
    """
    @classmethod
    def __new__(cls, *_, **kwargs) -> Union[ChatbotsMqSavePrompt,
                                            ChatbotsMqNewPrompt, 
                                            ChatbotsMqSubmindResponse]:
        message_text = kwargs.get("message_text") or kwargs.get("messageText") \
            or kwargs.get("shout")
        kwargs['message_text'] = message_text
        
        if message_text == CcaiControl.SAVE_PROMPT_RESULTS.value:
            return ChatbotsMqSavePrompt(**kwargs)
        elif message_text == CcaiControl.CREATE_PROMPT.value:
            return ChatbotsMqNewPrompt(**kwargs)
        else:
            return ChatbotsMqSubmindResponse(**kwargs)




class ChatbotsMqSubmindsState(MQContext):
    class SubmindState(BaseModel):
        submind_id: str = Field(description="Connected submind's user_id")
        status: SubmindStatus = Field(
            description="Subminds's status in a particular conversation")

    subminds_per_cid: Dict[str, List[SubmindState]] = Field(
        description="List of submind participants per conversation ID")
    connected_subminds: Dict[str, ConnectedSubmind] = Field(
        description="Dict of submind `user_id` to `ConnectedSubmind` object")
    cid_submind_bans: Dict[str, List[str]] = Field(
        description="Dict of `cid` to list of banned submind `user_id`s")
    banned_subminds: List[str] = Field(
        description="List of globally banned submind `user_id`s")

    msg_type: Literal["subminds_state"] = Field(
        "subminds_state", description="Message type for SIO", deprecated=True)


class ChatbotsMqConfiguredPersonasRequest(MQContext):
    service_name: str = Field(
        description="Name of the service to get personas for")
    user_id: Optional[str] = Field(
        default=None, description="Optional user_id making with the request.")


class ChatbotsMqConfiguredPersonasResponse(MQContext):
    update_time: datetime = Field(
        description="Time the personas were last checked")
    items: List[LLMPersona] = Field(
        description="List of configured personas from Klat")

    context: dict = Field(deprecated=True)

    @model_validator(mode='before')
    @classmethod
    def validate_context(cls, values):
        # Deprecated context handling for backwards-compat.
        if 'context' not in values and 'message_id' in values:
            values['context'] = {"mq": {"message_id": values['message_id']}}
        return values
        
    def model_dump(self, **kwargs):
        """
        Override model_dump to include 'persona_name' field for each item based 
        on its 'name' for backwards-compat. with Klat server
        """
        by_alias = {}
        if 'by_alias' not in kwargs:
            # `by_alias` to include `persona_name` in serialized `LLMPersona`s
            by_alias = super().model_dump(by_alias=True, **kwargs)
        
        return {**super().model_dump(**kwargs), **by_alias}

    @classmethod
    def from_persona_request(cls, data: dict,
                               request: ChatbotsMqConfiguredPersonasRequest):
        data["items"] = [item for item in data["items"]
                         if request.service_name in item["supported_llms"]]
        return cls(**data, message_id=request.message_id,
                   routing_key=request.routing_key)


class ChatbotsMqPromptsDataRequest(MQContext):
    """
    Convenience class. The message payload here is just `MQContext`.
    """


class ChatbotsMqPromptsDataResponse(MQContext):
    records: List[str] = Field(description="List of configured prompts")

    context: dict = Field(deprecated=True)

    @model_validator(mode='before')
    @classmethod
    def validate_context(cls, values):
        # Deprecated context handling for backwards-compat.
        if 'context' not in values and 'message_id' in values:
            values['context'] = {"mq": {"message_id": values['message_id']}}
        return values
    
    @classmethod
    def from_prompt_data_request(cls, data: dict,
                               request: ChatbotsMqPromptsDataRequest):
        return cls(**data, message_id=request.message_id,
                   routing_key=request.routing_key)


class ChatbotsMqSubmindConnection(MQContext):
    user_id: str = Field(description="User ID of the submind", alias="nick")
    time: datetime = Field(
        default= datetime.now(tz=timezone.utc),
        description="Timestamp when the submind last connected")
    cids: Optional[List[str]] = Field(
        default=None, description="List of conversation IDs the submind is in")
    context: Optional[ConnectedSubmind] = Field(
        default=None,
        description="ConnectedSubmind definition of the connecting bot")

    @model_validator(mode='before')
    @classmethod
    def validate_context(cls, values):
        if "context" in values and isinstance(values["context"], dict):
            user_id = values.get("user_id") or values.get("nick") or ""
            values["context"].setdefault("service_name",
                                          user_id.rsplit('-',1)[0])
        return values


class ChatbotsMqSubmindDisconnection(MQContext):
    user_id: str = Field(description="User ID of the submind", alias="nick")


class ChatbotsMqSubmindInvitation(MQContext):
    cid: str = Field(description="Conversation ID to invite subminds to")
    requested_participants: List[str] = Field(
        description="List of submind User IDs to invite to the conversation")


class ChatbotsMqUpdateParticipatingSubminds(MQContext):
    cid: str = Field(description="Conversation ID to update")
    subminds_to_invite: List[str] = Field(
        default=[],
        description="List of submind User IDs to invite to the conversation")
    subminds_to_kick: List[str] = Field(
        default=[],
        description="List of submind User IDs to evict from the conversation")


class ChatbotsMqSubmindConversationBan(MQContext):
    user_id: str = Field(description="User ID of the submind", alias="nick")
    cid: str = Field(description="Conversation ID to (un)ban submind from")


class ChatbotsMqSubmindGlobalBan(MQContext):
    user_id: str = Field(description="User ID of the submind", alias="nick")


class ChatbotsMqSubmindResponseError(MQContext):
    message: Optional[str] = Field(default=None, alias="msg",
                                    description="Error message")


__all__ = [ChatbotsMqRequest.__name__, ChatbotsMqResponse.__name__,
           ChatbotsMqSubmindResponse.__name__, ChatbotsMqSavePrompt.__name__,
           ChatbotsMqNewPrompt.__name__, ChatbotsMqSubmindsState.__name__, 
           ChatbotsMqConfiguredPersonasRequest.__name__,
           ChatbotsMqConfiguredPersonasResponse.__name__,
           ChatbotsMqPromptsDataRequest.__name__,
           ChatbotsMqPromptsDataResponse.__name__,
           ChatbotsMqSubmindConnection.__name__,
           ChatbotsMqSubmindDisconnection.__name__,
           ChatbotsMqSubmindInvitation.__name__,
           ChatbotsMqUpdateParticipatingSubminds.__name__,
           ChatbotsMqSubmindConversationBan.__name__,
           ChatbotsMqSubmindGlobalBan.__name__,
           ChatbotsMqSubmindResponseError.__name__,]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2024 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import List, Optional, Any, Dict, Literal, Union
from pydantic import Field

from neon_data_models.models.base.contexts import MQContext
from neon_data_models.models.api.llm import BrainForgeLLM, LLMRequest, LLMResponse, LLMPersona
from neon_data_models.models.api.http.brainforge import LLMGetModelsHttpResponse, LLMGetPersonasHttpRequest


class LLMGetModels(MQContext):
    user_id: str = Field(
        description="ID of user to get models for")


class LLMGetModelsResponse(MQContext, LLMGetModelsHttpResponse):
    pass


class LLMGetPersonas(LLMGetModels, LLMGetPersonasHttpRequest):
    @property
    def model_name(self):
        return self.model_id.split("@")[0]

    @property
    def model_version(self):
        return self.model_id.split("@", 1)[1]


class LLMGetPersonasResponse(MQContext):
    model: Optional[BrainForgeLLM] = Field(
        "Full configuration of requested model if model is loaded and access "
        "is allowed, else None.")

    @property
    def personas(self) -> List[LLMPersona]:
        """
        Convenience property defined to easily reference the personas requested
        """
        return self.model.personas if self.model else []


class LLMGetInference(LLMRequest, MQContext):
    user_id: str = Field("ID of user making the request")

    def as_llm_request(self):
        """
        Get a plain `LLMRequest` object from this `LLMGetInference` object.
        """
        return LLMRequest(**self.model_dump())


class LLMGetCompletion(LLMGetModels):
    model: str = Field(description="Model to request (<name>@<revision>)")
    completion_kwargs: Dict[str, Any] = Field(
        description="Dictionary of kwargs to pass to OpenAI completion request")


class LLMGetCompletionResponse(MQContext):
    openai_response: dict = Field(
        description="OpenAI ChatCompletion model")


class LLMGetTokenizerChatTemplatedString(LLMGetModels):
    model: str = Field(description="Model to request (<name>@<revision>)")
    messages: List[Dict[Literal["role", "content"], str]] = Field(
        description="List of dict messages in OpenAI format")
    add_generation_prompt: bool = Field(
        description="If true, assistant start tokens will be appended to the "
                    "formatted output.")
    tokenize: bool = Field(
        False,
        description="If true, a list of token strings is returned, "
                    "else a single string")


class LLMGetTokenizerChatTemplatedStringResponse(MQContext):
    prompt: Union[List[str], str] = Field(
        description="Prompt generated by the tokenizer")


class LLMGetInferenceResponse(LLMResponse, MQContext):
    pass


__all__ = [LLMGetModels.__name__, LLMGetModelsResponse.__name__,
           LLMGetPersonas.__name__, LLMGetPersonasResponse.__name__,
           LLMGetInference.__name__, LLMGetInferenceResponse.__name__,
           LLMGetCompletion.__name__, LLMGetCompletionResponse.__name__,
           LLMGetTokenizerChatTemplatedString.__name__,
           LLMGetTokenizerChatTemplatedStringResponse.__name__]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Annotated, List, Literal, Union
from ovos_bus_client.message import Message
from pydantic import Field, TypeAdapter, model_validator

from neon_data_models.models.base import BaseModel
from neon_data_models.models.base.contexts import (
    KlatContext,
    MQContext,
    SessionContext,
    TimingContext,
)
from neon_data_models.models.base.messagebus import BaseMessage, MessageContext
from neon_data_models.models.user.neon_profile import UserProfile
from neon_data_models.models.api.messagebus import (
    NeonGetLanguages,
    NeonGetTts,
    NeonGetStt,
    NeonAudioInput,
    NeonLanguagesResponse,
    NeonTextInput,
    NeonSttResponse,
    NeonTtsResponse,
    NeonGetSkillsApi,
    NeonSkillsApiResponse,
    NeonCallSkillApi,
    NeonCallSkillApiResponse,
    SkillApiRequestData
)


class GetTtsData(BaseModel):
    text: str = Field(description="Text to be spoken")
    lang: str = Field(
        default="en-us", description="BCP-47 language code for TTS"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_inputs(cls, values):
        if "text" not in values:
            values["text"] = values.pop("utterance", None)
        return values


class GetSttData(BaseModel):
    audio_data: str = Field(description="Base64-encoded audio data")
    lang: str = Field(
        default="en-us", description="BCP-47 language code for STT"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_inputs(cls, values):
        if "audio_data" not in values:
            values["audio_data"] = values.get("message_body")
        return values


class GetResponseData(BaseModel):
    utterances: List[str] = Field(description="List of input utterance(s)")
    lang: str = Field(
        default="en-us", description="BCP-47 language code for input/response"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_inputs(cls, values):
        if "utterances" not in values:
            values["utterances"] = [values.pop("messageText", "")]
        return values


class NeonMqGetTts(NeonGetTts, MQContext):
    """
    Data model for an MQ message requesting TTS synthesis
    """


class NeonMqGetStt(NeonGetStt, MQContext):
    """
    Data model for an MQ message requesting STT recognition
    """


class NeonMqTextInput(NeonTextInput, MQContext):
    """
    Data model for an MQ message containing text input
    """


class NeonMqAudioInput(NeonAudioInput, MQContext):
    """
    Data model for an MQ message containing audio input
    """


class NeonMqSttResponse(NeonSttResponse, MQContext):
    """
    Data model for an MQ message containing an STT recognition response
    """


class NeonMqTtsResponse(NeonTtsResponse, MQContext):
    """
    Data model for an MQ message containing a TTS synthesis response
    """


class NeonMqGetLanguages(NeonGetLanguages, MQContext):
    """
    Data model for an MQ message requesting supported languages
    """


class NeonMqLanguagesResponse(NeonLanguagesResponse, MQContext):
    """
    Data model for an MQ message containing supported languages response
    """


class NeonMqGetSkillsApi(NeonGetSkillsApi, MQContext):
    """
    Data model for an MQ message requesting available skill APIs
    """


class NeonMqSkillsApiResponse(NeonSkillsApiResponse, MQContext):
    """
    Data model for an MQ message containing available skill APIs
    """


class NeonMqCallSkillApi(NeonCallSkillApi, MQContext):
    """
    Data model for an MQ message calling a skill API
    """

    class _RequestData(SkillApiRequestData):
        msg_type: str = Field(
            description="Message type associated with the specific API requested"
        )

    msg_type: Literal["neon.skill_api.call"] = Field(
        default="neon.skill_api.call",
        description="Message type for skill API calls",
    )
    data: _RequestData = Field(
        description="API request data including `msg_type` and call params"
    )

    def as_messagebus_message(self) -> Message:
        """
        Override default Message translation to account for `msg_type` being
        specified in request data
        """
        return Message(
            msg_type=self.data.msg_type,
            data={"args": self.data.args, "kwargs": self.data.kwargs},
            context=self.context.model_dump()
            )

class NeonMqCallSkillApiResponse(NeonCallSkillApiResponse, MQContext):
    """
    Data model for an MQ message containing a skill API response.
    Note that `msg_type` is overridden from the internal API-specific value
    to a generic type for MQ routing.
    """
    msg_type: Literal["neon.skill_api.response"] = Field(
        default="neon.skill_api.response",
        description="Message type for skill API responses",
    )


class NeonMqUnknownMessage(BaseMessage, MQContext):
    """
    Default message class for validating Messagebus messages that should be
    forwarded to an MQ service
    """


class NeonApiMessage:
    """
    Type adapter for validating an arbitrary MQ message. This will always return
    an instance that extends `BaseMessage` and `MQContext`.
    """

    ta = TypeAdapter(
        Annotated[
            Union[
                NeonMqGetStt,
                NeonMqGetTts,
                NeonMqTextInput,
                NeonMqSttResponse,
                NeonMqTtsResponse,
                NeonMqGetLanguages,
                NeonMqLanguagesResponse,
                NeonMqGetSkillsApi,
                NeonMqSkillsApiResponse,
                NeonMqCallSkillApi,
                NeonMqCallSkillApiResponse,
            ],
            Field(discriminator="msg_type"),
        ]
    )

    @classmethod
    def __new__(cls, *args, **kwargs) -> BaseMessage:
        # Parse the MQ Context from a `Message` input to create a proper API message
        if "message_id" not in kwargs:
            # Extract MQ context data from the message
            mq_data = kwargs.get("context", {}).get("mq", {})
            # Update values with MQ context data
            kwargs.update(mq_data)

        try:
            return cls.ta.validate_python(kwargs)
        except Exception as e:
            # If validation fails, use the default message object
            return NeonMqUnknownMessage(**kwargs)

    @staticmethod
    def from_sio_message(sio_message: dict) -> BaseMessage:
        """
        Parse a Klat SocketIO message into a valid MQ API message
        """
        requested_service = sio_message.get(
            "requested_skill", "recognizer"
        ).lower()
        if requested_service not in ["stt", "tts", "recognizer"]:
            raise ValueError(
                f"Invalid requested service '{requested_service}'"
            )
        klat_context = KlatContext(**sio_message)
        mq_context = MQContext(**sio_message)
        context = MessageContext(
            source="mq_api",
            client=sio_message.get("client", "unknown"),
            username=sio_message.get("nick", "guest"),
            klat_data=klat_context,
            mq=mq_context,
            user_profiles=[UserProfile()],
            session=SessionContext(session_id=sio_message.get("cid", "klat")),
            timing=TimingContext(client_sent=sio_message.get("timeCreated")),
        )
        if requested_service == "stt":
            context.destination = ["speech"]
            return NeonMqGetStt(
                data=GetSttData(**sio_message),
                context=context,
                **mq_context.model_dump(),
            )
        elif requested_service == "tts":
            context.destination = ["audio"]
            return NeonMqGetTts(
                data=GetTtsData(**sio_message),
                context=context,
                **mq_context.model_dump(),
            )
        elif requested_service == "recognizer":
            context.destination = ["skills"]
            return NeonMqTextInput(
                data=GetResponseData(**sio_message),
                context=context,
                **mq_context.model_dump(),
            )


__all__ = [
    NeonMqGetTts.__name__,
    NeonMqGetStt.__name__,
    NeonMqTextInput.__name__,
    NeonMqAudioInput.__name__,
    NeonMqSttResponse.__name__,
    NeonMqTtsResponse.__name__,
    NeonMqGetLanguages.__name__,
    NeonMqLanguagesResponse.__name__,
    NeonApiMessage.__name__,
    NeonMqGetSkillsApi.__name__,
    NeonMqSkillsApiResponse.__name__,
    NeonMqCallSkillApi.__name__,
    NeonMqCallSkillApiResponse.__name__,
    NeonMqUnknownMessage.__name__,
]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Literal, Optional, Annotated, Union
from pydantic import Field, TypeAdapter, model_validator

from neon_data_models.models.api.jwt import HanaToken
from neon_data_models.models.base.contexts import MQContext
from neon_data_models.models.user.database import User


class CreateUserRequest(MQContext):
    operation: Literal["create"] = "create"
    user: User = Field(description="User object to create")


class ReadUserRequest(MQContext):
    operation: Literal["read"] = "read"
    user_spec: str = Field(description="Username or User ID to read")
    auth_user_spec: str = Field(
        default="", description="Username or ID to authorize database  read. "
                                "If unset, this will use `user_spec`")
    access_token: Optional[HanaToken] = Field(
        None, description="Token associated with `auth_username`")
    password: Optional[str] = Field(None,
                                    description="Password associated with "
                                                "`auth_username`")

    @model_validator(mode="after")
    def validate_params(self) -> 'ReadUserRequest':
        if not self.auth_user_spec:
            self.auth_user_spec = self.user_spec
        if self.access_token and self.access_token.purpose != "access":
            raise ValueError(f"Expected an access token but got: "
                             f"{self.access_token.purpose}")
        return self


class UpdateUserRequest(MQContext):
    operation: Literal["update"] = "update"
    user: User = Field(description="Updated User object to write to database")
    auth_username: str = Field(
        default="", description="Username to authorize database change. If "
                                "unset, this will use `user.username`")
    auth_password: str = Field(
        default="", description="Password (clear or hashed) associated with "
                                "`auth_username`. If unset, this will use "
                                "`user.password_hash`. If changing the "
                                "password, this must contain the existing "
                                "password, with the new password specified in "
                                "`user`")

    @model_validator(mode="after")
    def get_auth_username(self) -> 'UpdateUserRequest':
        if not self.auth_username:
            self.auth_username = self.user.username
        if not self.auth_password:
            self.auth_password = self.user.password_hash
        if not all((self.auth_username, self.auth_password)):
            raise ValueError("Missing auth_username or auth_password")
        return self


class DeleteUserRequest(MQContext):
    operation: Literal["delete"] = "delete"
    user: User = Field(description="Exact User object to remove from the "
                                   "database")


class UserDbRequest:
    """
    Generic class to dynamically build a UserDB CRUD request object based on the
    requested `operation`
    """
    ta = TypeAdapter(Annotated[Union[CreateUserRequest, ReadUserRequest,
                                     UpdateUserRequest, DeleteUserRequest],
                               Field(discriminator='operation')])

    def __new__(cls, *args, **kwargs):
        return cls.ta.validate_python(kwargs)


__all__ = [CreateUserRequest.__name__, ReadUserRequest.__name__,
           UpdateUserRequest.__name__, DeleteUserRequest.__name__,
           UserDbRequest.__name__]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime, timedelta
from pydantic import Field
from typing import List, Literal, Optional, Annotated, Dict

from neon_data_models.enum import UserData, AlertType, Weekdays
from neon_data_models.models.base import BaseModel
from neon_data_models.models.base.messagebus import BaseMessage, MessageContext


"""
This module contains models for interacting via the Node socket (WS).
"""


class AudioInputData(BaseModel):
    audio_data: str = Field(description="base64-encoded audio")
    lang: str = Field(description="BCP-47 language code")


class KlatResponse(BaseModel):
    sentence: str = Field(description="Text response")
    audio: Dict[Literal["male", "female"], Optional[str]] = Field(
        description="Mapping of gender to b64-encoded audio")


class AudioInputResponseData(BaseModel):
    parser_data: dict
    transcripts: List[str]
    skills_recv: bool


class NodeAudioInput(BaseMessage):
    msg_type: Literal["neon.audio_input"] = "neon.audio_input"
    data: AudioInputData


class NodeTextInput(BaseMessage):
    class UtteranceInputData(BaseModel):
        utterances: List[str] = Field(description="List of input utterance(s)")
        lang: str = Field(description="BCP-47 language")

    msg_type: Literal["recognizer_loop:utterance"] = "recognizer_loop:utterance"
    data: UtteranceInputData


class NodeGetStt(BaseMessage):
    msg_type: Literal["neon.get_stt"] = "neon.get_stt"
    data: AudioInputData


class NodeGetTts(BaseMessage):
    class TextInputData(BaseModel):
        text: str = Field(description="String text input")
        lang: str = Field(description="BCP-47 language code")

    msg_type: Literal["neon.get_tts"] = "neon.get_tts"
    data: TextInputData


class NodeKlatResponse(BaseMessage):
    msg_type: Literal["klat.response"] = "klat.response"
    data: Dict[str, KlatResponse] = Field(
        description="dict of BCP-47 language: KlatResponse")


class NodeAudioInputResponse(BaseMessage):
    msg_type: Literal["neon.audio_input.response"] = "neon.audio_input.response"
    data: AudioInputResponseData


class NodeGetSttResponse(BaseMessage):
    msg_type: Literal["neon.get_stt.response"] = "neon.get_stt.response"
    data: AudioInputResponseData


class NodeGetTtsResponse(BaseMessage):
    msg_type: Literal["neon.get_tts.response"] = "neon.get_tts.response"
    data: Dict[str, KlatResponse] = (
        Field(description="dict of BCP-47 language: KlatResponse"))


class CoreWWDetected(BaseMessage):
    msg_type: Literal["neon.ww_detected"] = "neon.ww_detected"
    # TODO: Define/implement schema in neon-speech service
    data: dict


class CoreIntentFailure(BaseMessage):
    msg_type: Literal["complete_intent_failure"] = "complete_intent_failure"
    data: dict  # Empty dict


class CoreErrorResponse(BaseMessage):
    class KlatErrorData(BaseModel):
        error: str = "unknown error"
        data: dict = {}

    msg_type: Literal["klat.error"] = "klat.error"
    data: KlatErrorData


class CoreClearData(BaseMessage):
    class ClearDataData(BaseModel):
        username: str
        data_to_remove: List[UserData]

    msg_type: Literal["neon.clear_data"] = "neon.clear_data"
    data: ClearDataData


class CoreAlertExpired(BaseMessage):
    class AlertData(BaseModel):
        alert_type: AlertType
        priority: Annotated[int, Field(gt=1, lt=10)]
        alert_name: str
        context: MessageContext
        next_expiration_time: Optional[datetime]
        repeat_frequency: Optional[timedelta]
        repeat_days: Optional[List[Weekdays]]
        end_repeat: Optional[datetime]
        audio_file: Optional[str] = None
        script_filename: Optional[str] = None

    msg_type: Literal["neon.alert_expired"] = "neon.alert_expired"
    data: AlertData


__all__ = [NodeAudioInput.__name__, NodeTextInput.__name__, NodeGetStt.__name__,
           NodeGetTts.__name__, NodeKlatResponse.__name__,
           NodeAudioInputResponse.__name__, NodeGetSttResponse.__name__,
           NodeGetTtsResponse.__name__, CoreWWDetected.__name__,
           CoreIntentFailure.__name__, CoreErrorResponse.__name__,
           CoreClearData.__name__, CoreAlertExpired.__name__]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import List
from pydantic import Field, RootModel
from neon_data_models.models.api.messagebus import (
    NeonSkillApiData,
    SkillApiRequestData,
    SkillApiResponseData,
)


class NeonSkillApiHttpData(NeonSkillApiData):
    skill_id: str = Field(
        description="ID of the skill providing this API method"
    )
    api_method: str = Field(description="API method name")


class NeonHttpListSkillApiResponse(RootModel):
    root: List[NeonSkillApiHttpData]

    model_config = {
        "json_schema_extra": {
            "examples": [
                [
                    {
                        "help": "\n        API Method to build a list of examples as listed in skill metadata.\n        ",
                        "request_schema": None,
                        "response_schema": None,
                        "signature": None,
                        "type": "skill-about.neongeckocom.skill_info_examples",
                        "skill_id": "skill-about.neongeckocom",
                        "api_method": "skill_info_examples",
                    },
                    {
                        "help": "\n        Get the current timestamp in seconds since epoch.\n        :param request: Request containing location to get time of\n        :returns: Response containing current timestamp\n        ",
                        "request_schema": {
                            "properties": {
                                "location": {
                                    "anyOf": [
                                        {"type": "string"},
                                        {"type": "null"},
                                    ],
                                    "default": None,
                                    "title": "Location",
                                }
                            },
                            "title": "_CurrentTimeRequest",
                            "type": "object",
                        },
                        "response_schema": {
                            "properties": {
                                "current_timestamp": {
                                    "title": "Current Timestamp",
                                    "type": "number",
                                }
                            },
                            "required": ["current_timestamp"],
                            "title": "_CurrentTimeResponse",
                            "type": "object",
                        },
                        "signature": "(request: skill_date_time._CurrentTimeRequest) -> skill_date_time._CurrentTimeResponse",
                        "type": "skill-date_time.neongeckocom.get_current_time",
                        "skill_id": "skill-date_time.neongeckocom",
                        "api_method": "get_current_time",
                    },
                ]
            ]
        }
    }


class NeonHttpSkillApiRequest(SkillApiRequestData):
    skill_id: str = Field(description="skill_id being requested")
    api_method: str = Field(description="API method being requested")

    model_config = {
        "json_schema_extra": {
            "examples": [
                {
                    "skill_id": "skill-about.neongeckocom",
                    "api_method": "skill_info_examples",
                    "args": [],
                    "kwargs": {},
                }
            ]
        }
    }


class NeonHttpSkillApiResponse(SkillApiResponseData):
    """
    Convenience wrapper to implement examples for Swagger UI
    """

    model_config = {
        "json_schema_extra": {
            "examples": [
                {
                    "result": "API Response matching advertised schema",
                    "error": None,
                },
                {"result": None, "error": "API Method error message"},
            ]
        }
    }


__all__ = [
    NeonHttpListSkillApiResponse.__name__,
    NeonHttpSkillApiRequest.__name__,
    NeonHttpSkillApiResponse.__name__,
]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from time import time
from typing import Optional, List, Literal
from uuid import uuid4

from pydantic import Field
from neon_data_models.enum import AccessRoles
from neon_data_models.models.base import BaseModel


class JWT(BaseModel):
    iss: Optional[str] = Field(None, validate_default=True,
                               description="Token issuer")
    sub: Optional[str] = Field(None, validate_default=True,
                               description="Unique token subject, ie a user ID")
    exp: int = Field(description="Expiration time in epoch seconds")
    iat: int = Field(description="Token creation time in epoch seconds")
    jti: str = Field(description="Unique token identifier",
                     default_factory=lambda: str(uuid4()))

    client_id: str = Field(description="Client identifier")
    roles: List[str] = Field(description="List of roles, "
                                         "formatted as `<name> <AccessRole>`. "
                                         "See PermissionsConfig for role names")


class HanaToken(JWT):
    def __init__(self, **kwargs):
        from neon_data_models.models.user import PermissionsConfig
        permissions = kwargs.get("permissions")
        if permissions and isinstance(permissions, PermissionsConfig):
            kwargs["roles"] = permissions.to_roles()
        elif permissions and isinstance(permissions, dict):
            core_permissions = AccessRoles.GUEST if \
                permissions.get("assist") else AccessRoles.NONE
            diana_permissions = AccessRoles.GUEST if \
                permissions.get("backend") else AccessRoles.NONE
            node_permissions = AccessRoles.USER if \
                permissions.get("node") else AccessRoles.NONE
            kwargs["roles"] = [f"core {core_permissions.value}",
                               f"diana {diana_permissions.value}",
                               f"node {node_permissions.value}"]
        if kwargs.get("expire") and isinstance(kwargs["expire"], float):
            kwargs["expire"] = round(kwargs["expire"])
        BaseModel.__init__(self, **kwargs)

    # Private parameters
    token_name: str = Field(default="",
                            description="Friendly name to identify this token.")
    creation_timestamp: int = Field(default_factory=lambda: int(time()),
                                    description="Timestamp of initial token "
                                                "creation (not counting "
                                                "refreshes).")
    last_refresh_timestamp: Optional[int] = Field(default=None,
                                                  description="Timestamp of "
                                                              "most recent "
                                                              "refresh.")
    purpose: Literal["access", "refresh"] = "access"


__all__ = [JWT.__name__, HanaToken.__name__]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from enum import Enum, IntEnum


class AccessRoles(IntEnum):
    """
    Defines access roles such that a larger value always corresponds to more
    permissions. `0` equates to no permission, negative numbers correspond to
    non-user roles. In this way, an activity can require, for example,
    `permission > AccessRoles.GUEST` to grant access to all registered users,
    admins, and owners.

    Special Roles:
        NODE: Reserved for use by a Node device service account to access
            various services
    """
    NONE = 0
    # 1-9 reserved for unauthenticated connections
    GUEST = 10
    # 11-19 reserved for unverified user roles
    USER = 20
    # 21-29 Reserved for "premium users"
    ADMIN = 30
    # 31-39 Reserved for "restricted owners"
    OWNER = 40
    # 41-49 Reserved for "escalated owners"

    # 50 Reserved for "unlimited access"

    NODE = -1


class UserData(IntEnum):
    """
    Defines types of user data.
    """
    CACHES = 0
    PROFILE = 1
    TRANSCRIPTS = 2
    LIKED_BRANDS = 3
    DISLIKED_BRANDS = 4
    ALL_DATA = 5
    ALL_MEDIA = 6
    UNITS_CONFIG = 7
    LANGUAGE_CONFIG = 8


class AlertType(IntEnum):
    """
    Defines kinds of alerts.
    """
    ALL = -1
    ALARM = 0
    TIMER = 1
    REMINDER = 2
    UNKNOWN = 99


class Weekdays(IntEnum):
    """
    Defines weekdays as used in the Alerts skill.
    """
    MON = 0
    TUE = 1
    WED = 2
    THU = 3
    FRI = 4
    SAT = 5
    SUN = 6


class SubmindStatus(Enum):
    """
    Defines status of a submind in a particular conversation
    """
    BANNED = "banned"
    ACTIVE = "active"


class CcaiState(IntEnum):
    """
    Defines status of a conversation where a CCAI is operating
    """
    IDLE = 0  # No active prompt
    RESP = 1  # Gathering responses to prompt
    DISC = 2  # Discussing responses
    VOTE = 3  # Voting on responses
    PICK = 4  # Proctor will select response
    WAIT = 5  # Bot is waiting for an invitation to participate


class CcaiControl(Enum):
    """
    Defines control messages for a CCAI conversation
    """
    PROMPT_REGEX = r"^!PROMPT:"
    SAVE_PROMPT_RESULTS = "!MSG:SAVE_PROMPT_RESULTS"
    CREATE_PROMPT = "!MSG:CREATE_PROMPT"
    START_AUTO_PROMPTS = "!START_AUTO_PROMPTS"
    STOP_AUTO_PROMPTS = "!STOP_AUTO_PROMPTS"



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Optional, Dict, List
from pydantic import Field

from neon_data_models.models.api.llm import LLMRequest, LLMPersona
from neon_data_models.models.base.contexts import MQContext


class LLMProposeRequest(MQContext, LLMRequest):
    model: Optional[str] = Field(
        default=None,
        description="MQ implementation defines `model` as optional because the "
                    "queue defines the requested model in most cases.")
    persona: Optional[LLMPersona] = Field(
        default=None,
        description="MQ implementation defines `persona` as an optional "
                    "parameter, with default behavior hard-coded into each "
                    "LLM module.")


class LLMProposeResponse(MQContext):
    response: str = Field(description="LLM response to the prompt")


class LLMDiscussRequest(LLMProposeRequest):
    options: Dict[str, str] = Field(
        description="Mapping of participant name to response to be discussed.")


class LLMDiscussResponse(MQContext):
    opinion: str = Field(description="LLM response to the available options.")


class LLMVoteRequest(LLMProposeRequest):
    responses: List[str] = Field(
        description="List of responses to choose from.")


class LLMVoteResponse(MQContext):
    sorted_answer_indexes: List[int] = Field(
        description="Indices of `responses` ordered high to low by preference.")


__all__ = [LLMProposeRequest.__name__, LLMProposeResponse.__name__,
           LLMDiscussRequest.__name__, LLMDiscussResponse.__name__,
           LLMVoteRequest.__name__, LLMVoteResponse.__name__]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytz
import datetime

from typing import Optional, List, Literal

from pydantic import Field

from neon_data_models.models.base import BaseModel

from neon_data_models.models.user.database import User


class ProfileUser(BaseModel):
    first_name: str = ""
    middle_name: str = ""
    last_name: str = ""
    preferred_name: str = ""
    full_name: str = ""
    dob: str = "YYYY/MM/DD"
    age: str = ""
    email: str = ""
    username: str = ""
    password: str = ""
    picture: str = Field(default="",
                         description="Fully-qualified URI of a user avatar. "
                                     "(i.e. `https://example.com/avatar.jpg")
    about: str = ""
    phone: str = ""
    phone_verified: bool = False
    email_verified: bool = False


class ProfileSpeech(BaseModel):
    stt_language: str = "en-us"
    alt_languages: List[str] = ['en']
    tts_language: str = "en-us"
    tts_gender: str = "female"
    neon_voice: Optional[str] = ''
    secondary_tts_language: Optional[str] = ''
    secondary_tts_gender: str = "male"
    secondary_neon_voice: str = ''
    speed_multiplier: float = 1.0


class ProfileUnits(BaseModel):
    time: Literal[12, 24] = 12
    date: Literal["MDY", "YMD", "YDM", "DMY"] = "MDY"
    measure: Literal["imperial", "metric"] = "imperial"


class ProfileLocation(BaseModel):
    lat: Optional[float] = None
    lng: Optional[float] = None
    city: Optional[str] = None
    state: Optional[str] = None
    country: Optional[str] = None
    tz: Optional[str] = None
    utc: Optional[float] = None


class ProfileResponseMode(BaseModel):
    speed_mode: str = "quick"
    hesitation: bool = False
    limit_dialog: bool = False


class ProfilePrivacy(BaseModel):
    save_audio: bool = False
    save_text: bool = False


class UserProfile(BaseModel):
    user: ProfileUser = ProfileUser()
    speech: ProfileSpeech = ProfileSpeech()
    units: ProfileUnits = ProfileUnits()
    location: ProfileLocation = ProfileLocation()
    response_mode: ProfileResponseMode = ProfileResponseMode()
    privacy: ProfilePrivacy = ProfilePrivacy()

    @classmethod
    def from_user_object(cls, user: User):
        user_config = user.neon
        today = datetime.date.today()
        if user_config.user.dob:
            dob = user_config.user.dob
            age = str(today.year - dob.year - (
                    (today.month, today.day) < (dob.month, dob.day)))
            dob = dob.strftime("%Y/%m/%d")
        else:
            age = ""
            dob = "YYYY/MM/DD"
        full_name = " ".join((n for n in (user_config.user.first_name,
                                          user_config.user.middle_name,
                                          user_config.user.last_name) if n))
        user = ProfileUser(about=user_config.user.about,
                           age=age, dob=dob,
                           email=user_config.user.email,
                           email_verified=False,
                           first_name=user_config.user.first_name,
                           full_name=full_name,
                           last_name=user_config.user.last_name,
                           middle_name=user_config.user.middle_name,
                           password=user.password_hash or "",
                           phone=user_config.user.phone,
                           phone_verified=False,
                           picture=user_config.user.avatar_url,
                           preferred_name=user_config.user.preferred_name,
                           username=user.username
                           )
        alt_stt = [lang.split('-')[0] for lang in
                   user_config.language.input_languages[1:]]
        secondary_tts_lang = user_config.language.output_languages[1] if (
                len(user_config.language.output_languages) > 1) else None
        speech = ProfileSpeech(
            alt_languages=alt_stt,
            secondary_tts_gender=user_config.response_mode.tts_gender,
            secondary_tts_language=secondary_tts_lang,
            speed_multiplier=user_config.response_mode.tts_speed_multiplier,
            stt_language=user_config.language.input_languages[0].split('-')[0],
            tts_gender=user_config.response_mode.tts_gender,
            tts_language=user_config.language.output_languages[0])
        units = ProfileUnits(**user_config.units.model_dump())
        utc_hours = (pytz.timezone(user_config.location.timezone or "UTC")
                     .utcoffset(datetime.datetime.now()).total_seconds() / 3600)
        # TODO: Get city, state, country from lat/lon
        location = ProfileLocation(lat=user_config.location.latitude,
                                   lng=user_config.location.longitude,
                                   tz=user_config.location.timezone,
                                   utc=utc_hours)
        response_mode = ProfileResponseMode(
            **user_config.response_mode.model_dump())
        privacy = ProfilePrivacy(**user_config.privacy.model_dump())

        return UserProfile(location=location, privacy=privacy,
                           response_mode=response_mode, speech=speech,
                           units=units, user=user)


__all__ = [ProfileUser.__name__, ProfileSpeech.__name__, ProfileUnits.__name__,
           ProfileLocation.__name__, ProfileResponseMode.__name__,
           ProfilePrivacy.__name__, UserProfile.__name__]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Defines models for `Message` objects sent on the Neon messagebus.
"""

from typing import Any, List, Literal, Dict, Optional
from pydantic import Field, model_validator

from neon_data_models.types import Gender
from neon_data_models.models.base import BaseModel
from neon_data_models.models.base.messagebus import BaseMessage


# Data models
class GetTtsData(BaseModel):
    text: str = Field(description="Text to be spoken")
    lang: str = Field(
        default="en-us", description="BCP-47 language code for TTS"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_inputs(cls, values):
        # Ensure input values are normalized to valid field names
        if hasattr(values, "model_dump"):
            values = values.model_dump()
        if "text" not in values:
            values["text"] = values.get("utterance")
        return values


class TtsResponse(BaseModel):
    sentence: str = Field(description="Text to be spoken")
    translated: bool = Field(
        default=False,
        description="True if the text has been translated before TTS synthesis",
    )
    phonemes: Optional[str] = Field(
        default=None, description="Phoneme representation of the sentence"
    )
    genders: List[Gender] = Field(
        description="List of genders included in the response `audio` field",
        deprecated=True,
    )
    male: Optional[str] = Field(
        default=None, description="Path to audio file in male voice"
    )
    female: Optional[str] = Field(
        default=None, description="Path to audio file in female voice"
    )
    audio: Dict[Gender, str]

    @model_validator(mode="after")
    def validate_gender_audio_match(self):
        """Validate that the genders list matches the audio dictionary keys."""
        if not set(self.genders) == set(self.audio.keys()):
            raise ValueError(
                "All genders listed in 'genders' must have corresponding "
                "entries in 'audio'"
            )
        return self


class TtsSpeaker(BaseModel):
    name: str = Field(default="Neon", description="Name of the speaker")
    speaker: Optional[str] = Field(
        default=None,
        description="Deprecated: Use 'name' instead",
        deprecated=True,
    )
    language: str = Field(default="en-us", description="BCP-47 language code")
    gender: Gender = Field(
        default="female", description="Requested or synthesized gender"
    )
    voice: Optional[str] = Field(
        default=None, description="Requested or synthesized voice name"
    )

    @model_validator(mode="before")
    @classmethod
    def map_speaker_to_name(cls, values):
        if hasattr(values, "model_dump"):
            values = values.model_dump()
        if "name" not in values and "speaker" in values:
            values["name"] = values["speaker"]
        return values


class TtsReponseData(BaseModel):
    responses: Dict[str, TtsResponse] = Field(
        description="Dictionary of BCP-47 lang codes to TTS responses"
    )
    speaker: Optional[TtsSpeaker] = Field(
        default=None, description="Optional speaker metadata"
    )


class GetSttData(BaseModel):
    audio_data: str = Field(description="Base64-encoded audio data")
    lang: str = Field(
        default="en-us", description="BCP-47 language code for STT"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_inputs(cls, values):
        # Ensure input values are normalized to valid field names
        if hasattr(values, "model_dump"):
            values = values.model_dump()
        if "audio_data" not in values:
            values["audio_data"] = values.get("message_body")
        return values


class SttReponseData(BaseModel):
    transcripts: List[str]
    parser_data: Dict[str, Any]

    @model_validator(mode="after")
    def validate_transcripts_not_empty(self):
        """Validate that the transcripts list is not empty."""
        if not self.transcripts:
            raise ValueError("'transcripts' cannot be an empty list")
        return self


class GetResponseData(BaseModel):
    utterances: List[str] = Field(description="List of input utterance(s)")
    lang: str = Field(
        default="en-us", description="BCP-47 language code for input/response"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_inputs(cls, values):
        # Ensure input values are normalized to valid field names
        if hasattr(values, "model_dump"):
            values = values.model_dump()
        if "utterances" not in values:
            values["utterances"] = [values.pop("messageText", "")]
        return values


class NeonLanguagesData(BaseModel):
    stt: List[str] = Field(description="List of supported STT languages")
    tts: List[str] = Field(description="List of supported TTS languages")
    skills: List[str] = Field(description="List of supported skill languages")


# Message models
class NeonGetTts(BaseMessage):
    msg_type: Literal["neon.get_tts"] = "neon.get_tts"
    data: GetTtsData

    @model_validator(mode="after")
    def ensure_audio_destination(self):
        if "audio" not in self.context.destination:
            self.context.destination = ["audio"]
        return self


class NeonGetStt(BaseMessage):
    msg_type: Literal["neon.get_stt"] = "neon.get_stt"
    data: GetSttData

    @model_validator(mode="after")
    def ensure_audio_destination(self):
        if "audio" not in self.context.destination:
            # Default context is not valid for this request, so override it
            self.context.destination = ["audio"]
        return self


class NeonTextInput(BaseMessage):
    msg_type: Literal["recognizer_loop:utterance"] = (
        "recognizer_loop:utterance"
    )
    data: GetResponseData

    @model_validator(mode="after")
    def ensure_skills_destination(self):
        if "skills" not in self.context.destination:
            self.context.destination.append("skills")
        return self


class NeonAudioInput(BaseMessage):
    msg_type: Literal["neon.audio_input"] = "neon.audio_input"
    data: GetSttData

    @model_validator(mode="after")
    def ensure_audio_destination(self):
        if "audio" not in self.context.destination:
            # Default context is not valid for this request, so override it
            self.context.destination = ["audio"]
        return self


class NeonSttResponse(BaseMessage):
    msg_type: Literal["neon.get_stt.response"] = "neon.get_stt.response"
    data: SttReponseData


class NeonTtsResponse(BaseMessage):
    msg_type: Literal["neon.get_tts.response", "klat.response"] = (
        "neon.get_tts.response"
    )
    data: TtsReponseData


class NeonGetLanguages(BaseMessage):
    msg_type: Literal["neon.languages.get"] = "neon.languages.get"


class NeonLanguagesResponse(BaseMessage):
    msg_type: Literal["neon.languages.get.response"] = (
        "neon.languages.get.response"
    )
    data: NeonLanguagesData


class NeonGetSkillsApi(BaseMessage):
    msg_type: Literal["neon.skill_api.get"] = "neon.skill_api.get"


class NeonSkillApiData(BaseModel):
    help: str = Field(description="API Method docstring")
    request_schema: Optional[dict] = Field(
        default=None, description="JSON request schema for the API method"
    )
    response_schema: Optional[dict] = Field(
        default=None, description="JSON response schema for the API method"
    )
    signature: Optional[str] = Field(
        default=None, description="Python signature of the API method"
    )
    msg_type: str = Field(
        alias="type",
        description="Message type associated with this API method",
    )


class NeonSkillsApiResponse(BaseMessage):
    msg_type: Literal["neon.skill_api.get.response"] = (
        "neon.skill_api.get.response"
    )
    data: Dict[str, Dict[str, NeonSkillApiData]] = Field(
        description="Dict of skill_id to API methods"
    )


class SkillApiRequestData(BaseModel):
    args: List[Any] = Field(default=[], description="Positional arguments")
    kwargs: Dict[str, Any] = Field(default={}, description="Keyword arguments")


class SkillApiResponseData(BaseModel):
    result: Any = Field(
        description="Result of the API method call",
        default=None,
    )
    error: Optional[str] = Field(
        default=None, description="Error message if the call failed"
    )


class NeonCallSkillApi(BaseMessage):
    msg_type: str = Field(description="Requested API Message type")
    data: SkillApiRequestData = Field(
        default=SkillApiRequestData(),
        description="Data for the API method call",
    )


class NeonCallSkillApiResponse(BaseMessage):
    msg_type: str = Field(description="Requested API response Message type")
    data: SkillApiResponseData = Field(description="API Method response data")


__all__ = [
    NeonGetTts.__name__,
    NeonGetStt.__name__,
    NeonTextInput.__name__,
    NeonAudioInput.__name__,
    NeonSttResponse.__name__,
    NeonTtsResponse.__name__,
    NeonGetLanguages.__name__,
    NeonLanguagesResponse.__name__,
    NeonGetSkillsApi.__name__,
    NeonSkillsApiResponse.__name__,
    NeonCallSkillApi.__name__,
    NeonCallSkillApiResponse.__name__,
]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from enum import Enum
from os import environ
from datetime import datetime, timedelta
from pydantic import ConfigDict, BaseModel as _BaseModel


class BaseModel(_BaseModel):
    model_config = ConfigDict(extra="allow" if environ.get(
            "NEON_DATA_MODELS_ALLOW_EXTRA", "false") != "false" else "ignore",
            populate_by_name=environ.get("NEON_DATA_MODELS_POPULATE_BY_NAME",
                                          "true") != "false",)
        
    def model_dump(self, *args, **kwargs) -> dict:
        """
        Global `model_dump` overrides to ensure model serialization.
        Recursively processes nested dictionaries, lists, and BaseModel instances.
        """
        data = super().model_dump(*args, **kwargs)
        return self._process_data(data)
    
    def _process_data(self, data):
        """
        Recursively process data to convert datetime and timedelta objects.
        """
        if isinstance(data, dict):
            # Process dictionary values
            for key, value in list(data.items()):
                data[key] = self._process_data(value)
            return data
        elif isinstance(data, list):
            # Process list elements
            return [self._process_data(item) for item in data]
        elif isinstance(data, datetime):
            # Convert datetime to timestamp
            return data.timestamp()
        elif isinstance(data, timedelta):
            # Convert timedelta to seconds
            return data.total_seconds()
        elif isinstance(data, Enum):
            # Always serialize `Enum` objects by value
            return data.value
        else:
            # Return other types unchanged
            return data





# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime, timedelta, timezone
from typing import Literal, List, Optional, Any, Tuple
from uuid import uuid4

from pydantic import Field, model_validator

from neon_data_models.models.base import BaseModel


class SessionContext(BaseModel):
    session_id: str = "default"
    active_skills: List[Tuple[str, float]] = Field(
        default=[], 
        description="List of tuple `skill_id` and last used timestamp")
    utterance_states: dict = {}
    lang: Optional[str] = None
    context: dict = {}
    site_id: str = "unknown"
    pipeline: List[str] = []
    location: dict = {}
    system_unit: Optional[Literal["imperial", "metric"]] = None
    date_format: Optional[Literal["MDY", "YMD", "YDM", "DMY"]] = None
    time: Optional[Literal[12, 24]] = None
    is_recording: bool = False
    is_speaking: bool = False
    blacklisted_skills: List[str] = []
    blacklisted_intents: List[str] = []

    def model_dump(self, *args, **kwargs) -> dict:
        # Override to explicitly exclude default `None` values so that upstream
        # logic works to read values from global config
        kwargs["exclude_none"] = True
        return BaseModel.model_dump(self, *args, **kwargs)


class TimingContext(BaseModel):
    def __init__(self, **kwargs):
        # Enables backwards-compat. with old context values
        if transcribed := kwargs.pop("transcribed", None):
            kwargs.setdefault("handle_utterance", transcribed)
        if text_parsers := kwargs.pop("text_parsers", None):
            kwargs.setdefault("transform_utterance", text_parsers)
        BaseModel.__init__(self, **kwargs)

    audio_begin: Optional[datetime] = None
    audio_end: Optional[datetime] = None
    client_sent: Optional[datetime] = None
    gradio_sent: Optional[datetime] = None
    handle_utterance: Optional[datetime] = None
    response_sent: Optional[datetime] = None
    speech_start: Optional[datetime] = None

    get_stt: Optional[timedelta] = None
    get_tts: Optional[timedelta] = None
    iris_input_handling: Optional[timedelta] = None
    mq_response_handler: Optional[timedelta] = None
    mq_from_core: Optional[timedelta] = None
    mq_from_client: Optional[timedelta] = None
    mq_input_handler: Optional[timedelta] = None
    client_to_core: Optional[timedelta] = None
    client_from_core: Optional[timedelta] = None
    save_transcript: Optional[timedelta] = None
    transform_audio: Optional[timedelta] = None
    transform_utterance: Optional[timedelta] = None
    wait_in_queue: Optional[timedelta] = None

    @model_validator(mode='before')
    @classmethod
    def validate_datetime_fields(cls, values):
        """
        Validate datetime fields to parse epoch timestamps into datetime objects.
        Dynamically identifies fields with datetime type annotations.
        """
        # Find all fields that have datetime type annotation
        datetime_fields = [
            field_name for field_name, field_info in cls.model_fields.items()
            if field_info.annotation == datetime or (
                hasattr(field_info.annotation, "__origin__") and 
                field_info.annotation.__origin__ is Optional and 
                datetime in field_info.annotation.__args__
            )
        ]
        
        for field in datetime_fields:
            if field in values and values[field] is not None:
                if isinstance(values[field], (int, float)):
                    # Convert epoch timestamp to datetime with UTC timezone
                    values[field] = datetime.fromtimestamp(values[field], tz=timezone.utc)
                elif not isinstance(values[field], datetime):
                    # If it's neither a timestamp nor a datetime, raise an error
                    raise ValueError(f"Field '{field}' must be a timestamp or datetime object")
        
        return values


class KlatContext(BaseModel):
    sid: str = Field(default="", description="Klat Shout ID")
    cid: str = Field(default="", description="Klat Conversation ID")
    title: str = Field(default="", description="Klat Conversation Title")

    @model_validator(mode='before')
    @classmethod
    def validate_inputs(cls, values):
        """
        Validate KlatContext inputs to normalize messageID to sid.
        """
        values.setdefault("sid", values.get("messageID", ""))
        if "title" in values and values["title"] is None:
            values.pop("title")
        return values

class MQContext(BaseModel):
    routing_key: Optional[str] = None
    message_id: str = Field(default_factory=lambda: uuid4().hex,
                            description="MQ unique message ID")

    @model_validator(mode='before')
    @classmethod
    def validate_inputs(cls, values):
        """
        Validate MQContext inputs to normalize messageID to message_id.
        """
        values.setdefault("message_id", values.get("messageID"))
        if values['message_id'] is None:
            # Allow default_factory to generate a message_id
            values.pop("message_id")
        return values


class GradioContext(BaseModel):
    session: str = Field(description="Gradio session ID")



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import List, Tuple, Optional, Literal
from pydantic import Field, model_validator, computed_field

from neon_data_models.models.base import BaseModel


_DEFAULT_MQ_TO_ROLE = {"user": "user", "llm": "assistant"}


class LLMPersonaIdentity(BaseModel):
    """
    Defines metadata for a unique persona.
    """
    name: str = Field(alias="persona_name", 
                      description="Unique name for this persona")
    user_id: Optional[str] = Field(
        None, description="`user_id` of the user who created this persona.")

    @computed_field
    @property
    def id(self) -> str:
        persona_id = self.name
        if self.user_id:
            persona_id += f"_{self.user_id}"
        return persona_id


class LLMPersona(LLMPersonaIdentity):
    """
    Complete persona definition that may be applied to LLM inference or
    committed to a database.
    """
    description: Optional[str] = Field(
        None, description="Human-readable description of this persona")
    system_prompt: Optional[str] = Field(
        None, description="System prompt associated with this persona. "
                          "If None, `description` will be used.")
    enabled: bool = Field(
        True, description="Flag used to mark a defined persona as "
                          "available for use.")

    @model_validator(mode='after')
    def validate_request(self):
        if self.name == "vanilla":
            assert self.system_prompt in (None, "")
            self.system_prompt = None
            return self

        assert any(x is not None for x in (self.description, self.system_prompt))
        if self.system_prompt is None:
            self.system_prompt = self.description
        return self


class LLMRequest(BaseModel):
    query: str = Field(description="Incoming user prompt")
    # TODO: History may support more options in the future
    history: List[Tuple[Literal["user", "llm"], str]] = Field(
        description="Formatted chat history (excluding system prompt). Note "
                    "that the roles used here will differ from those used in "
                    "OpenAI-compatible requests.")
    persona: LLMPersona = Field(
        description="Requested persona to respond to this message")
    model: str = Field(description="Model to request (<name>@<revision>)")
    max_tokens: int = Field(
        default=512, ge=64, le=2048,
        description="Maximum number of tokens to include in the response")
    temperature: float = Field(
        default=0.0, ge=0.0, le=1.0,
        description="Temperature of response. 0 guarantees reproducibility, "
                    "higher values increase variability. Must be `0.0` if "
                    "`beam_search` is True")
    repetition_penalty: float = Field(
        default=1.0, ge=1.0, le=2.0,
        description="Repetition penalty. Higher values limit repeated "
                    "information in responses")
    stream: bool = Field(
        default=None, description="Enable streaming responses. "
                                  "Mutually exclusive with `beam_search`.")
    best_of: int = Field(
        default=1, ge=1,
        description="Number of beams to use if `beam_search` is enabled.")
    beam_search: bool = Field(
        default=None, description="Enable beam search. "
                                  "Mutually exclusive with `stream`.")
    max_history: int = Field(
        default=2, description="Maximum number of user/assistant "
                               "message pairs to include in history context. "
                               "Excludes system prompt and incoming query.")

    @model_validator(mode='before')
    @classmethod
    def validate_inputs(cls, values):
        # Neon modules previously defined `user` and `llm` keys, but Open AI
        # specifies `assistant` in place of `llm` and is the de-facto standard
        for idx, itm in enumerate(values.get('history', [])):
            if itm[0] == "assistant":
                values['history'][idx] = ("llm", itm[1])
        # OpenAI `extra_body` may be included in input; parse those inputs
        if values.get('use_beam_search') is not None:
            values['beam_search'] = values['use_beam_search']
        return values

    @model_validator(mode='after')
    def validate_request(self):
        # If beams are specified, make sure valid `stream` and `beam_search`
        # values are specified
        if self.best_of > 1:
            if self.stream is True:
                raise ValueError("Cannot stream with a `best_of` value "
                                 "greater than 1")
            if self.beam_search is False:
                raise ValueError("Cannot have a `best_of` value other than 1 "
                                 "if `beam_search` is False")
            self.stream = False
            self.beam_search = True
        # If streaming, beam_search must be False
        if self.stream is True:
            if self.beam_search is True:
                raise ValueError("Cannot enable both `stream` and "
                                 "`beam_search`")
            self.beam_search = False
        # If beam search is enabled, `best_of` must be >1
        if self.beam_search is True and self.best_of <= 1:
            raise ValueError(f"best_of must be greater than 1 when using "
                             f"beam search. Got {self.best_of}")
        # If beam search is enabled, streaming must be False
        if self.beam_search is True:
            if self.stream is True:
                raise ValueError("Cannot enable both `stream` and "
                                 "`beam_search`")
            self.stream = False
        if self.stream is None and self.beam_search is None:
            self.stream = True
            self.beam_search = False

        assert isinstance(self.stream, bool)
        assert isinstance(self.beam_search, bool)

        # If beam search is enabled, temperature must be set to 0.0
        if self.beam_search:
            assert self.temperature == 0.0, "Beam search requires temperature 0"
        return self

    @property
    def messages(self) -> List[dict]:
        """
        Get chat history as a list of dict messages
        """
        return [{"role": m[0], "content": m[1]} for m in self.history]

    def to_completion_kwargs(self, mq2role: dict = None) -> dict:
        """
        Get kwargs to pass to an OpenAI completion request.
        @param mq2role: dict mapping `llm` and `user` keys to `role` values to
            use in message history.
        """
        mq2role = mq2role or _DEFAULT_MQ_TO_ROLE
        history = self.messages[-2*self.max_history:]
        for msg in history:
            msg["role"] = mq2role.get(msg["role"]) or msg["role"]
        if self.persona.system_prompt is not None:
            history.insert(0, {"role": "system",
                               "content": self.persona.system_prompt})
        history.append({"role": "user", "content": self.query})
        return {"model": self.model,
                "messages": history,
                "max_tokens": self.max_tokens,
                "temperature": self.temperature,
                "stream": self.stream,
                "extra_body": {"add_special_tokens": True,
                               "repetition_penalty": self.repetition_penalty,
                               "use_beam_search": self.beam_search,
                               "best_of": self.best_of}}


class LLMResponse(BaseModel):
    response: str = Field(description="LLM Response to the input query")
    history: List[Tuple[Literal["user", "llm"], str]] = Field(
        description="List of (role, content) tuples in chronological order "
                    "(`response` is in the last list element)")
    finish_reason: Literal["length", "stop"] = Field(
        "stop", description="Reason response generation ended.")

    @model_validator(mode='before')
    @classmethod
    def validate_inputs(cls, values):
        # Neon modules previously defined `user` and `llm` keys, but Open AI
        # specifies `assistant` in place of `llm` and is the de-facto standard
        for idx, itm in enumerate(values.get('history', [])):
            if itm[0] == "assistant":
                values['history'][idx] = ("llm", itm[1])
        return values


class BrainForgeLLM(BaseModel):
    name: str = Field(description="LLM Name")
    version: str = Field(description="LLM Version")
    personas: List[LLMPersona] = Field(
        default=[], description="List of personas defined in this model")

    @property
    def vllm_spec(self):
        """
        Model identifier used by vllm (<name>@<version>)
        """
        return f"{self.name}@{self.version}"


__all__ = [LLMPersonaIdentity.__name__, LLMPersona.__name__,
           LLMRequest.__name__, LLMResponse.__name__,
           BrainForgeLLM.__name__]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from time import time
from typing import Dict, Any, List, Literal, Optional
from typing_extensions import deprecated
from uuid import uuid4

# from neon_data_models.models.api import HanaToken
from neon_data_models.models.base import BaseModel
from pydantic import Field, field_validator
from datetime import date

from neon_data_models.enum import AccessRoles


class _UserConfig(BaseModel):
    first_name: str = ""
    middle_name: str = ""
    last_name: str = ""
    preferred_name: str = ""
    dob: Optional[date] = None
    email: str = ""
    avatar_url: str = Field(default="",
                            description="Fully-qualified URI of a user avatar. "
                                        "(i.e. `https://example.com/avatar.jpg")
    about: str = ""
    phone: str = ""
    
    @field_validator('dob', mode='before')
    @classmethod
    def validate_dob(cls, v):
        if v == "YYYY/MM/DD":
            # Legacy default value for dob should be treated as `None`
            return None
        return v


class _LanguageConfig(BaseModel):
    input_languages: List[str] = ["en-us"]
    output_languages: List[str] = ["en-us"]


class _UnitsConfig(BaseModel):
    time: Literal[12, 24] = 12
    date: Literal["MDY", "YMD", "YDM", "DMY"] = "MDY"
    measure: Literal["imperial", "metric"] = "imperial"


class _ResponseConfig(BaseModel):
    hesitation: bool = False
    limit_dialog: bool = False
    tts_gender: Literal["male", "female"] = "female"
    tts_speed_multiplier: float = 1.0


class _LocationConfig(BaseModel):
    latitude: Optional[float] = None
    longitude: Optional[float] = None
    name: Optional[str] = None
    timezone: Optional[str] = None


class _PrivacyConfig(BaseModel):
    save_text: bool = True
    save_audio: bool = False


class NeonUserConfig(BaseModel):
    """
    Defines user configuration used in Neon Core.
    """
    skills: Dict[str, Dict[str, Any]] = {}
    user: _UserConfig = _UserConfig()
    # Former `speech` schema is replaced by `language` which is a more general
    # format.
    language: _LanguageConfig = _LanguageConfig()
    units: _UnitsConfig = _UnitsConfig()
    # Former `location` schema is replaced here with a minimal spec from which
    # the remaining values may be calculated
    location: _LocationConfig = _LocationConfig()
    response_mode: _ResponseConfig = _ResponseConfig()
    privacy: _PrivacyConfig = _PrivacyConfig()


class KlatConfig(BaseModel):
    """
    Defines user configuration used in PyKlatChat.
    """
    is_tmp: bool = True
    preferences: Dict[str, Any] = {}


class BrainForgeConfig(BaseModel):
    """
    Defines configuration used in BrainForge LLM applications.
    """
    inference_access: Dict[str, Dict[str, List[str]]] = {}


class PermissionsConfig(BaseModel):
    """
    Defines roles for supported projects/services.
    """
    klat: AccessRoles = Field(
        AccessRoles.NONE, description="Defines access to Klat chat services.")
    core: AccessRoles = Field(
        AccessRoles.NONE, description="Defines access to Neon core services.")
    diana: AccessRoles = Field(
        AccessRoles.NONE,
        description="Defines access to DIANA backend services. "
                    "(i.e. API proxy, email proxy).")
    users: AccessRoles = Field(
        AccessRoles.NONE, description="Defines access to the users service.")
    node: AccessRoles = Field(
        AccessRoles.NONE,
        description="Defines access to the node websocket in HANA.")
    hub: AccessRoles = Field(
        AccessRoles.NONE, description="Defines access to a hub device.")
    llm: AccessRoles = Field(
        AccessRoles.NONE,
        description="Defines access to the BrainForge LLM backend. Note that "
                    "per-model permissions may also apply and further restrict "
                    "a user's access to some models for inference.")

    class Config:
        use_enum_values = True
        validate_default = True

    @classmethod
    def from_roles(cls, roles: List[str]):
        """
        Parse PermissionsConfig from standard JWT roles configuration.
        """
        kwargs = {}
        for role in roles:
            name, value = role.split(' ')
            kwargs[name] = AccessRoles[value.upper()]
        return cls(**kwargs)

    def to_roles(self):
        """
        Dump a PermissionsConfig to standard JWT roles to be included in a JWT.
        """
        roles = []
        for key, val in self.model_dump().items():
            roles.append(f"{key} {AccessRoles(val).name}")
        return roles


@deprecated(f"Use `neon_data_models.models.api.jwt.HanaToken`")
class TokenConfig(BaseModel):
    username: str
    client_id: str
    permissions: Dict[str, bool]
    refresh_token: str
    expiration: int = Field(
        description="Unix timestamp of auth token expiration")
    refresh_expiration: int = Field(
        description="Unix timestamp of refresh token expiration")
    token_name: str
    creation_timestamp: int = Field(
        description="Unix timestamp of token creation (auth+refresh)")
    last_refresh_timestamp: int = Field(
        description="Unix timestamp of last token refresh (auth+refresh)")
    access_token: Optional[str] = None


class User(BaseModel):

    @classmethod
    def rebuild_model(cls):
        # Ensure `HanaToken` is populated from the import space
        from neon_data_models.models.api.jwt import HanaToken
        cls.model_rebuild()

    username: str
    password_hash: Optional[str] = None
    user_id: str = Field(default_factory=lambda: str(uuid4()))
    created_timestamp: int = Field(default_factory=lambda: round(time()))
    neon: NeonUserConfig = NeonUserConfig()
    klat: KlatConfig = KlatConfig()
    llm: BrainForgeConfig = BrainForgeConfig()
    permissions: PermissionsConfig = PermissionsConfig()
    tokens: Optional[List['HanaToken']] = [] # type: ignore

    def __eq__(self, other):
        return self.model_dump() == other.model_dump()


__all__ = [NeonUserConfig.__name__, KlatConfig.__name__,
           BrainForgeConfig.__name__, PermissionsConfig.__name__, User.__name__]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Optional, List, Union
from pydantic import ConfigDict, Field, field_validator, model_validator

from neon_data_models.models.base import BaseModel
from neon_data_models.models.base.contexts import (GradioContext, 
                                                   SessionContext, KlatContext,
                                                   TimingContext, MQContext)
from neon_data_models.models.client.node import NodeData
from neon_data_models.models.user.neon_profile import UserProfile


class MessageContext(BaseModel):
    model_config = ConfigDict(extra="allow")
    session: SessionContext = Field(description="Session Data",
                                              default=SessionContext())
    node_data: Optional[NodeData] = Field(description="Node Data", default=None)
    timing: TimingContext = Field(
        description="User Interaction Timing Information", 
        default=TimingContext())
    user_profiles: Optional[List[UserProfile]] = (
        Field(description="List of relevant user profiles", default=None))
    klat_data: Optional[KlatContext] = Field(
        description="Klat context for Klat-generated messages", default=None)
    mq: Optional[MQContext] = Field(
        description="MQ context for messages traversing a RabbitMQ broker",
        default=None)

    username: str = "local"
    # TODO: Consider refactoring client/client_name into a single dict
    #  or merging with `node_data`
    client_name: str = "unknown"
    client: str = "unknown"
    source: Union[str, List[str]] = "unknown"
    destination: Union[str, List[str]] = Field(
        default=["skills"],
        description="List of destination modules expected to handle the message"
        )
    neon_should_respond: bool = True
    gradio: Optional[GradioContext] = Field(
        default=None,
        description="Context for messages originating from a Gradio interface")
    
    @field_validator('destination')
    def validate_destination(cls, v):
        if isinstance(v, str):
            return [v]
        return v
    
    @model_validator(mode='before')
    def validate_session_value(cls, data):
        if isinstance(data, dict) and data.get('session', {}) is None:
            data.pop('session')
        return data


class BaseMessage(BaseModel):
    msg_type: str
    data: dict
    context: MessageContext

    def as_messagebus_message(self) -> "Message": # type: ignore
        """
        Get a `Message` representation of this object.
        """
        try:
            from ovos_bus_client.message import Message
            if hasattr(self.data, 'model_dump'):
                data = self.data.model_dump()
            else:
                data = self.data
            return Message(self.msg_type, data, self.context.model_dump())
        except ImportError:
            raise RuntimeError("pip install ovos-bus-client to enable Message "
                               "deserialization.")



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2024 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import List, Optional, Dict, Literal, Any
from uuid import uuid4
from time import time
from pydantic import Field, model_validator

from neon_data_models.models.base import BaseModel
from neon_data_models.models.api.llm import (
    BrainForgeLLM,
    LLMPersona,
    LLMRequest,
    LLMResponse,
)


class LLMGetModelsHttpResponse(BaseModel):
    models: List[BrainForgeLLM]


class LLMGetPersonasHttpRequest(BaseModel):
    model_id: str = Field(
        description="Model ID (<name>@<version>) to get personas for"
    )


class LLMGetPersonasHttpResponse(BaseModel):
    personas: List[LLMPersona] = Field(
        description="List of personas associated with the requested model."
    )


class LLMGetInferenceHttpRequest(LLMRequest):
    llm_name: str = Field(description="Model name to request")
    llm_revision: str = Field(description="Model revision to request")
    model: Optional[str] = Field(
        default=None,
        description="Model ID (<name>@<version>) used for vLLM inference",
    )

    @model_validator(mode="after")
    def set_model_from_name_and_revision(self):
        if self.model is None:
            self.model = f"{self.llm_name}@{self.llm_revision}"
        return self


class OpenAiCompletionRequest(BaseModel):
    model: str = Field(description="Model ID (generally <name>@<version>)")
    messages: List[Dict[Literal["role", "content"], str]] = Field(
        description="List of messages to send to the model for completion, "
        "including `system` and other context messages."
    )
    max_tokens: Optional[int] = Field(
        default=512, description="Maximum number of tokens to generate"
    )
    temperature: Optional[float] = Field(
        default=0.0, description="Temperature of generation"
    )
    stream: Optional[bool] = Field(
        default=False, description="Not Implemented."
    )
    persona: Optional[LLMPersona] = Field(default=None)
    extra_body: Optional[Dict[str, Any]] = Field(default=None)

    model_config = {
        "json_schema_extra": {
            "examples": [
                {
                    "model": "<name>@<version>",
                    "messages": [
                        {
                            "role": "system",
                            "content": "You are a helpful assistant.",
                        },
                        {"role": "user", "content": "Who are you?"},
                    ],
                    "max_tokens": 512,
                    "temperature": 0.0,
                    "stream": False,
                    "extra_body": {
                        "use_beam_search": True,
                        "best_of": 3,
                    }
                }
            ]
        }
    }

    @model_validator(mode="after")
    def ensure_default_values(self):
        """Set default values for compat with BrainForge API."""
        self.max_tokens = self.max_tokens or 512
        self.temperature = self.temperature or 0.0
        if self.stream is None:
            self.stream = False
        if self.extra_body is None:
            self.extra_body = {}
        return self

    @model_validator(mode="after")
    def get_persona(self):
        """Determine a persona based on message history or default vanilla."""
        for message in self.messages:
            if message["role"] == "system":
                sys_message = self.messages.pop(self.messages.index(message))
                if sys_message["content"]:
                    self.persona = LLMPersona(
                        name="custom", system_prompt=sys_message["content"]
                    )
                    return self
                break
        self.persona = LLMPersona(name="vanilla", system_prompt=None)
        return self

    @model_validator(mode="after")
    def validate_messages(self):
        if len(self.messages) < 1:
            raise ValueError("At least one `user` message is required.")
        return self

    def to_llm_inference_http_request(self) -> LLMGetInferenceHttpRequest:
        """
        Convert this OpenAI completion request to LLMGetInferenceHttpRequest
        """
        model, revision = self.model.split("@", 1)
        query = self.messages.pop(-1)["content"]
        history = [(msg["role"], msg["content"]) for msg in self.messages]
        return LLMGetInferenceHttpRequest(
            llm_name=model,
            llm_revision=revision,
            model=self.model,
            max_tokens=self.max_tokens,
            temperature=self.temperature,
            query=query,
            history=history,
            persona=self.persona,
            **self.extra_body
        )


class OpenAiCompletionResponse(BaseModel):
    id: str = Field(
        description="UID for this completion request",
        default_factory=lambda: uuid4().hex,
    )
    object: Literal["chat.completion"] = "chat.completion"
    created: float = Field(
        description="Timestamp of completion request",
        default_factory=lambda: time(),
    )
    model: str = Field(description="Model ID used for this completion")
    choices: List[
        Dict[Literal["message"], Dict[Literal["role", "content"], str]]
    ] = Field(description="List of responses from the model")

    model_config = {
        "json_schema_extra": {
            "examples": [
                {
                    "id": uuid4().hex,
                    "object": "chat.completion",
                    "created": time(),
                    "model": "<name>@<version>",
                    "choices": [
                        {
                            "message": {
                                "role": "assistant",
                                "content": "This is a sample response",
                            }
                        }
                    ],
                }
            ]
        }
    }

    @classmethod
    def from_llm_response(
        cls, llm_response: LLMResponse, llm_request: OpenAiCompletionRequest
    ) -> "OpenAiCompletionResponse":
        """Get an OpenAI response from a BrainForge LLMResponse."""
        choices = [
            {
                "message": {
                    "role": "assistant",
                    "content": llm_response.response,
                }
            }
        ]
        return cls(model=llm_request.model, choices=choices)


__all__ = [
    LLMGetModelsHttpResponse.__name__,
    LLMGetPersonasHttpRequest.__name__,
    LLMGetPersonasHttpResponse.__name__,
    LLMGetInferenceHttpRequest.__name__,
    OpenAiCompletionRequest.__name__,
    OpenAiCompletionResponse.__name__,
]
